import pytest

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.tasks.metadata_etl.picklists import AddPicklistEntries
from cumulusci.utils.xml import metadata_tree

OBJECT_XML = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <fields>
        <fullName>Time_Zone__c</fullName>
        <type>Picklist</type>
        <valueSet>
            <valueSetDefinition>
                <sorted>false</sorted>
                <value>
                    <fullName>Pacific/Kiritimati</fullName>
                    <default>false</default>
                </value>
                <value>
                    <fullName>Pacific/Chatham</fullName>
                    <default>true</default>
                </value>
            </valueSetDefinition>
        </valueSet>
    </fields>
    <fields>
        <fullName>Type__c</fullName>
        <type>Picklist</type>
        <valueSet>
            <valueSetDefinition>
                <sorted>true</sorted>
                <value>
                    <fullName>Fundraising</fullName>
                    <default>false</default>
                </value>
                <value>
                    <fullName>Outreach</fullName>
                    <default>true</default>
                </value>
            </valueSetDefinition>
        </valueSet>
    </fields>
    <fields>
        <fullName>TestGVS__c</fullName>
        <externalId>false</externalId>
        <label>TestGVS</label>
        <required>false</required>
        <trackFeedHistory>false</trackFeedHistory>
        <type>Picklist</type>
        <valueSet>
            <restricted>true</restricted>
            <valueSetName>Test</valueSetName>
        </valueSet>
    </fields>
    <recordTypes>
        <fullName>Default_RT</fullName>
        <active>true</active>
        <picklistValues>
            <picklist>Type__c</picklist>
            <values>
                <fullName>Fundraising</fullName>
                <default>false</default>
            </values>
            <values>
                <fullName>Outreach</fullName>
                <default>true</default>
            </values>
        </picklistValues>
        <picklistValues>
            <picklist>Time_Zone__c</picklist>
            <values>
                <fullName>Pacific%2FKiritimati</fullName>
                <default>false</default>
            </values>
            <values>
                <fullName>Pacific%2FChatham</fullName>
                <default>false</default>
            </values>
        </picklistValues>
    </recordTypes>
    <recordTypes>
        <fullName>Second_RT</fullName>
        <active>true</active>
        <picklistValues>
            <picklist>Type__c</picklist>
            <values>
                <fullName>Fundraising</fullName>
                <default>false</default>
            </values>
            <values>
                <fullName>Outreach</fullName>
                <default>true</default>
            </values>
        </picklistValues>
        <picklistValues>
            <picklist>Time_Zone__c</picklist>
            <values>
                <fullName>Pacific%2FKiritimati</fullName>
                <default>false</default>
            </values>
            <values>
                <fullName>Pacific%2FChatham</fullName>
                <default>false</default>
            </values>
        </picklistValues>
    </recordTypes>
</CustomObject>
"""


class TestAddPicklistValues:
    def test_adds_values(self):
        task = create_task(
            AddPicklistEntries,
            {
                "api_version": "47.0",
                "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                "entries": [
                    {"fullName": "Test"},
                    {"fullName": "Foo", "label": "Bar", "default": True},
                ],
            },
        )

        # Validate that the first sObject has one picklist changed
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject")
        values = result.find(
            "fields", fullName="Time_Zone__c"
        ).valueSet.valueSetDefinition.value

        test_elem = next(v for v in values if v.fullName.text == "Test")
        assert test_elem is not None
        assert test_elem.label.text == "Test"
        assert test_elem.default.text == "false"

        test_elem = next(v for v in values if v.fullName.text == "Foo")
        assert test_elem is not None
        assert test_elem.label.text == "Bar"
        assert test_elem.default.text == "true"

        test_elem = next(v for v in values if v.fullName.text == "Pacific/Chatham")
        assert test_elem.default.text == "false"

        # Show that the other picklist is unaffected
        values = result.find(
            "fields", fullName="Type__c"
        ).valueSet.valueSetDefinition.value
        assert "Test" not in (v.fullName.text for v in values)

    def test_does_not_add_duplicate_values(self):
        task = create_task(
            AddPicklistEntries,
            {
                "api_version": "47.0",
                "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                "entries": [{"fullName": "Outreach"}],
            },
        )

        # Validate that the duplicate entry is not added
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject2")
        values = result.find(
            "fields", fullName="Type__c"
        ).valueSet.valueSetDefinition.value

        test_elem = list(v for v in values if v.fullName.text == "Outreach")
        assert len(test_elem) == 1

    def test_adds_values_second_object(self):
        task = create_task(
            AddPicklistEntries,
            {
                "api_version": "47.0",
                "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                "entries": [
                    {"fullName": "Test"},
                    {"fullName": "Foo", "label": "Bar", "default": True},
                ],
            },
        )

        # Validate that the second object has one picklist changed
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject2")
        values = result.find(
            "fields", fullName="Type__c"
        ).valueSet.valueSetDefinition.value

        test_elem = next(v for v in values if v.fullName.text == "Test")
        assert test_elem is not None
        assert test_elem.label.text == "Test"
        assert test_elem.default.text == "false"

        test_elem = next(v for v in values if v.fullName.text == "Foo")
        assert test_elem is not None
        assert test_elem.label.text == "Bar"
        assert test_elem.default.text == "true"

        test_elem = next(v for v in values if v.fullName.text == "Outreach")
        assert test_elem.default.text == "false"

        # Show that the other picklist is unaffected
        values = result.find(
            "fields", fullName="Time_Zone__c"
        ).valueSet.valueSetDefinition.value
        assert "Test" not in (v.fullName.text for v in values)

    def test_adds_record_type_entries(self):
        task = create_task(
            AddPicklistEntries,
            {
                "api_version": "47.0",
                "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                "entries": [
                    {"fullName": "Test"},
                    {"fullName": "Foo", "label": "Bar", "default": True},
                ],
                "record_types": ["*"],
            },
        )

        # Validate that the entries are added to the Record Type
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject2")

        for rt_name in ["Default_RT", "Second_RT"]:
            # Make sure we added the picklist values
            values = (
                result.find("recordTypes", fullName=rt_name)
                .find("picklistValues", picklist="Type__c")
                .values
            )
            assert "Test" in (v.fullName.text for v in values)
            assert "Foo" in (v.fullName.text for v in values)

            # Check that we set the default
            foo_value = next(v for v in values if v.fullName.text == "Foo")
            assert foo_value.default.text == "true"

            outreach_value = next(v for v in values if v.fullName.text == "Outreach")
            assert outreach_value.default.text == "false"

            # And that we did not add the other picklist's values
            values = (
                result.find("recordTypes", fullName=rt_name)
                .find("picklistValues", picklist="Time_Zone__c")
                .values
            )
            assert "Test" not in (v.fullName.text for v in values)

    def test_adds_record_type_entries__missing_picklist(self):
        task = create_task(
            AddPicklistEntries,
            {
                "api_version": "47.0",
                "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                "entries": [
                    {"fullName": "Test"},
                    {"fullName": "Foo", "label": "Bar", "default": True},
                ],
                "record_types": ["*"],
            },
        )

        # Validate that the entries are added to the Record Type
        # when the Record Type metadata doesn't already contain
        # an entry for this picklist
        tree = metadata_tree.fromstring(OBJECT_XML)
        tree.find("recordTypes", fullName="Default_RT").remove(
            tree.find("recordTypes", fullName="Default_RT").find(
                "picklistValues", picklist="Type__c"
            )
        )
        result = task._transform_entity(tree, "MyObject2")

        for rt_name in ["Default_RT", "Second_RT"]:
            # Make sure we added the picklist values
            values = (
                result.find("recordTypes", fullName=rt_name)
                .find("picklistValues", picklist="Type__c")
                .values
            )
            assert "Test" in (v.fullName.text for v in values)
            assert "Foo" in (v.fullName.text for v in values)

    def test_adds_record_type_entries__single(self):
        task = create_task(
            AddPicklistEntries,
            {
                "api_version": "47.0",
                "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                "entries": [
                    {"fullName": "Test"},
                    {"fullName": "Foo", "label": "Bar", "default": True},
                ],
                "record_types": ["Default_RT"],
            },
        )

        # Validate that the entries are added to the Record Type
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject2")

        values = (
            result.find("recordTypes", fullName="Default_RT")
            .find("picklistValues", picklist="Type__c")
            .values
        )
        assert "Test" in (v.fullName.text for v in values)
        assert "Foo" in (v.fullName.text for v in values)

        values = (
            result.find("recordTypes", fullName="Second_RT")
            .find("picklistValues", picklist="Type__c")
            .values
        )
        assert "Test" not in (v.fullName.text for v in values)

    def test_adds_record_type_entries__existing_quoted(self):
        task = create_task(
            AddPicklistEntries,
            {
                "api_version": "47.0",
                "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                "entries": [{"fullName": "Pacific%2FKiritimati"}],
                "record_types": ["Default_RT"],
            },
        )

        # Validate that the entries are not added to the Record Type in duplicate
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject")

        values = (
            result.find("recordTypes", fullName="Default_RT")
            .find("picklistValues", picklist="Time_Zone__c")
            .values
        )
        assert "Pacific%2FKiritimati" in (v.fullName.text for v in values)
        assert "Pacific/Kiritimati" not in (v.fullName.text for v in values)

    def test_adds_record_type_entries__multiple(self):
        task = create_task(
            AddPicklistEntries,
            {
                "api_version": "47.0",
                "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                "entries": [
                    {"fullName": "Test"},
                    {"fullName": "Foo", "label": "Bar", "default": True},
                ],
                "record_types": ["Default_RT", "Second_RT"],
            },
        )

        # Validate that the entries are added to the Record Type
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject2")

        for rt_name in ["Default_RT", "Second_RT"]:
            # Make sure we added the picklist values
            values = (
                result.find("recordTypes", fullName=rt_name)
                .find("picklistValues", picklist="Type__c")
                .values
            )
            assert "Test" in (v.fullName.text for v in values)
            assert "Foo" in (v.fullName.text for v in values)

    def test_add_before__existing(self):
        task = create_task(
            AddPicklistEntries,
            {
                "api_version": "47.0",
                "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                "entries": [
                    {"fullName": "Test", "add_before": "Pacific/Chatham"},
                    {"fullName": "Foo", "label": "Bar", "default": True},
                ],
                "record_types": ["Default_RT"],
            },
        )

        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject")
        vsd = result.find("fields", fullName="Time_Zone__c").valueSet.valueSetDefinition
        values = vsd.value
        test_elem = next(v for v in values if v.fullName.text == "Test")
        assert vsd._element.index(test_elem._element) == 2  # The `sorted` element is 0

        test_elem = next(v for v in values if v.fullName.text == "Foo")
        assert vsd._element.index(test_elem._element) == 4

        rt_picklist = result.find("recordTypes", fullName="Default_RT").find(
            "picklistValues", picklist="Time_Zone__c"
        )
        rt_values = rt_picklist.values
        test_elem = next(v for v in rt_values if v.fullName.text == "Test")
        assert (
            rt_picklist._element.index(test_elem._element) == 2
        )  # The `picklist` element is 0
        test_elem = next(v for v in rt_values if v.fullName.text == "Foo")
        assert rt_picklist._element.index(test_elem._element) == 4

    def test_add_before__missing(self):
        task = create_task(
            AddPicklistEntries,
            {
                "api_version": "47.0",
                "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                "entries": [
                    {"fullName": "Test", "add_before": "Not-there"},
                    {"fullName": "Foo", "label": "Bar", "default": True},
                ],
                "record_types": ["Default_RT"],
            },
        )

        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject")
        vsd = result.find("fields", fullName="Time_Zone__c").valueSet.valueSetDefinition
        values = vsd.value
        test_elem = next(v for v in values if v.fullName.text == "Test")
        assert vsd._element.index(test_elem._element) == 3  # The `sorted` element is 0

        test_elem = next(v for v in values if v.fullName.text == "Foo")
        assert vsd._element.index(test_elem._element) == 4

        rt_picklist = result.find("recordTypes", fullName="Default_RT").find(
            "picklistValues", picklist="Time_Zone__c"
        )
        rt_values = rt_picklist.values
        test_elem = next(v for v in rt_values if v.fullName.text == "Test")
        assert (
            rt_picklist._element.index(test_elem._element) == 3
        )  # The `picklist` element is 0
        test_elem = next(v for v in rt_values if v.fullName.text == "Foo")
        assert rt_picklist._element.index(test_elem._element) == 4

    def test_init_options__old_api_version(self):
        with pytest.raises(TaskOptionsError):
            create_task(
                AddPicklistEntries,
                {
                    "api_version": "35.0",
                    "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                    "entries": [{"fullName": "Test"}],
                },
            )

    def test_init_options__bad_api_version(self):
        with pytest.raises(TaskOptionsError):
            create_task(
                AddPicklistEntries,
                {
                    "api_version": "35q.0",
                    "picklists": ["MyObject.Time_Zone__c", "MyObject2.Type__c"],
                    "entries": [{"fullName": "Test"}],
                },
            )

    def test_init_options__bad_picklist_name(self):
        with pytest.raises(TaskOptionsError):
            create_task(
                AddPicklistEntries,
                {
                    "api_version": "47.0",
                    "picklists": ["MyObjectTime_Zone__c"],
                    "entries": [{"fullName": "Test"}],
                },
            )

    def test_init_options__standard_picklist_name(self):
        with pytest.raises(TaskOptionsError):
            create_task(
                AddPicklistEntries,
                {
                    "api_version": "47.0",
                    "picklists": ["Opportunity.StageName"],
                    "entries": [{"fullName": "Test"}],
                },
            )

    def test_init_options__missing_fullname(self):
        with pytest.raises(TaskOptionsError):
            create_task(
                AddPicklistEntries,
                {
                    "api_version": "47.0",
                    "picklists": ["Opportunity.Type__c"],
                    "entries": [{"label": "Test"}],
                },
            )

    def test_init_options__multiple_defaults(self):
        with pytest.raises(TaskOptionsError):
            create_task(
                AddPicklistEntries,
                {
                    "api_version": "47.0",
                    "picklists": ["Opportunity.Type__c"],
                    "entries": [
                        {"fullName": "Test", "default": True},
                        {"fullName": "Bar", "default": True},
                    ],
                },
            )

    def test_raises_for_missing_picklist(self):
        task = create_task(
            AddPicklistEntries,
            {
                "api_version": "47.0",
                "picklists": ["MyObject.Type2__c"],
                "entries": [{"fullName": "Test", "default": True}],
            },
        )
        tree = metadata_tree.fromstring(OBJECT_XML)
        with pytest.raises(TaskOptionsError):
            task._transform_entity(tree, "MyObject")

    def test_raises_for_missing_picklists_option(self):
        with pytest.raises(TaskOptionsError):
            create_task(
                AddPicklistEntries,
                {
                    "api_version": "47.0",
                    "entries": [{"fullName": "Test", "default": True}],
                },
            )

    def test_raises_for_global_value_set(self):
        with pytest.raises(TaskOptionsError):
            task = create_task(
                AddPicklistEntries,
                {
                    "api_version": "47.0",
                    "picklists": ["MyObject.TestGVS__c"],
                    "entries": [{"fullName": "Test", "default": True}],
                },
            )

            tree = metadata_tree.fromstring(OBJECT_XML)
            task._transform_entity(tree, "MyObject")



import csv
import io
from pathlib import PurePosixPath
from collections import defaultdict, namedtuple

from distutils.version import StrictVersion

from cumulusci.tasks.github.base import BaseGithubTask
from cumulusci.core.utils import process_bool_arg
from cumulusci.utils import download_extract_github_from_repo
from cumulusci.utils.xml import metadata_tree
from cumulusci.utils.yaml.cumulusci_yml import cci_safe_load
from cumulusci.core.exceptions import DependencyResolutionError, TaskOptionsError

Package = namedtuple("Package", ["repo", "package_name", "namespace", "prefix_release"])
PackageVersion = namedtuple("PackageVersion", ["package", "version"])
SObjectDetail = namedtuple(
    "SObjectDetail", ["version", "api_name", "label", "description"]
)

FieldDetail = namedtuple(
    "FieldDetail",
    [
        "version",
        "sobject",
        "api_name",
        "label",
        "type",
        "help_text",
        "description",
        "valid_values",
    ],
)


class GenerateDataDictionary(BaseGithubTask):
    task_docs = """
    Generate a data dictionary for the project by walking all GitHub releases.
    The data dictionary is output as two CSV files.
    One, in `object_path`, includes the Object Name, Object Label, and Version Introduced,
    with one row per packaged object.
    The other, in `field_path`, includes Object Name, Field Name, Field Label, Field Type,
    Valid Picklist Values (if any) or a Lookup referenced table (if any), Version Introduced.
    Both MDAPI and SFDX format releases are supported. However, only force-app/main/default
    is processed for SFDX projects.
    """

    task_options = {
        "object_path": {
            "description": "Path to a CSV file to contain an sObject-level data dictionary."
        },
        "field_path": {
            "description": "Path to a CSV file to contain an field-level data dictionary."
        },
        "include_dependencies": {
            "description": "Process all of the GitHub dependencies of this project and include their schema in the data dictionary.",
            "default": True,
        },
        "additional_dependencies": {
            "description": "Include schema from additional GitHub repositories that "
            "are not explicit dependencies of this project to build a unified data dictionary. "
            "Specify as a list of dicts as in project__dependencies in cumulusci.yml. Note: only "
            "repository dependencies are supported."
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        if self.options.get("object_path") is None:
            self.options[
                "object_path"
            ] = f"{self.project_config.project__name} Objects.csv"

        if self.options.get("field_path") is None:
            self.options[
                "field_path"
            ] = f"{self.project_config.project__name} Fields.csv"

        self.options["include_dependencies"] = process_bool_arg(
            self.options.get("include_dependencies", True)
        )

        if "additional_dependencies" in self.options and not all(
            ["github" in dep for dep in self.options["additional_dependencies"]]
        ):
            raise TaskOptionsError("Only GitHub dependencies are currently supported.")

    def _get_repo_dependencies(
        self, dependencies=None, include_beta=None, visited_repos=None
    ):
        """Return a list of Package objects representing all of the GitHub repositories
        in this project's dependency tree. Ignore all non-GitHub dependencies."""
        deps = []
        visited = visited_repos or set()

        for dependency in dependencies:
            if "github" in dependency:
                repo = self.project_config.get_repo_from_url(dependency["github"])
                if repo is None:
                    raise DependencyResolutionError(
                        f"Github repository {dependency['github']} not found or not authorized."
                    )
                _, ref = self.project_config.get_ref_for_dependency(
                    repo, dependency, include_beta
                )

                contents = repo.file_contents("cumulusci.yml", ref=ref)
                cumulusci_yml = cci_safe_load(
                    io.StringIO(contents.decoded.decode("utf-8"))
                )
                project = cumulusci_yml.get("project", {})
                namespace = project.get("package", {}).get("namespace", "")
                if namespace:
                    namespace = f"{namespace}__"
                else:
                    namespace = ""

                if f"{repo.owner}/{repo.name}" in visited:
                    continue

                deps.append(
                    Package(
                        repo,
                        project.get("package", {}).get(
                            "name", f"{repo.owner}/{repo.name}"
                        ),
                        namespace,
                        project.get("git", {}).get("prefix_release", "release/"),
                    )
                )
                visited.add(f"{repo.owner}/{repo.name}")

                deps.extend(
                    self._get_repo_dependencies(
                        cumulusci_yml.get("project", {}).get("dependencies", []),
                        visited_repos=visited,
                    )
                )

        return deps

    def _run_task(self):
        self.logger.info("Starting data dictionary generation")

        self._init_schema()

        # Find all of our dependencies, if we're processing dependencies.
        dependencies = self.project_config.project__dependencies or []
        namespace = self.project_config.project__package__namespace
        if namespace:
            namespace = f"{namespace}__"
        repos = [
            Package(
                self.get_repo(),
                self.project_config.project__package__name,
                namespace,
                self.project_config.project__git__prefix_release,
            )
        ]
        if "additional_dependencies" in self.options:
            repos += self._get_repo_dependencies(
                self.options["additional_dependencies"], include_beta=False
            )
        if self.options["include_dependencies"]:
            repos += self._get_repo_dependencies(dependencies, include_beta=False)

        for package in repos:
            self._walk_releases(package)

        self._write_results()

    def _init_schema(self):
        """Initialize the structure used for schema storage."""
        self.sobjects = defaultdict(list)
        self.fields = defaultdict(list)
        self.package_versions = defaultdict(list)
        self.omit_sobjects = set()

    def _walk_releases(self, package):
        """Traverse all of the releases in this project's repository and process
        each one matching our tag (not draft/prerelease) to generate the data dictionary."""
        for release in package.repo.releases():
            # Skip this release if any are true:
            # It is a draft release
            # It is prerelease (managed beta)
            # This release's tag does not have the expected prefix,
            # meaning we don't know its version number
            if (
                release.draft
                or release.prerelease
                or not release.tag_name.startswith(package.prefix_release)
            ):
                continue

            zip_file = download_extract_github_from_repo(
                package.repo, ref=release.tag_name
            )
            version = PackageVersion(
                package,
                self._version_from_tag_name(release.tag_name, package.prefix_release),
            )
            self.package_versions[package].append(version.version)
            self.logger.info(
                f"Analyzing {package.package_name} version {version.version}"
            )

            if "src/objects/" in zip_file.namelist():
                # MDAPI format
                self._process_mdapi_release(zip_file, version)

            if "force-app/main/default/objects/" in zip_file.namelist():
                # TODO: check sfdx-project.json for directories to process.
                # SFDX format
                self._process_sfdx_release(zip_file, version)

    def _process_mdapi_release(self, zip_file, version):
        """Process an MDAPI ZIP file for objects and fields"""
        for f in zip_file.namelist():
            path = PurePosixPath(f)
            if path.parent == PurePosixPath("src/objects") and path.suffix == ".object":
                sobject_name = path.stem
                if sobject_name.count("__") == 1:
                    sobject_name = f"{version.package.namespace}{sobject_name}"

                self._process_object_element(
                    sobject_name, metadata_tree.fromstring(zip_file.read(f)), version
                )

    def _should_process_object(self, namespace, sobject_name, element):
        """Determine if we should track this object in the object dictionary.
        Fields may be included regardless.

        We don't process custom objects owned by other namespaces.
        It's fine for Package A to package fields on an object owned by Package B.
        We'll record the fields on their owning package (B) and the object on its (A).

        We also do not process Custom Settings, Platform Events, or Custom Metadata Types."""

        return (
            (sobject_name.endswith("__c") or sobject_name.count("__") == 0)
            and sobject_name.startswith(namespace)
            and element.find("customSettingsType") is None
            if element
            else True
        )

    def _should_process_object_fields(self, sobject_name, element):
        """Determine if we should track fields from this object in the field dictionary.
        The object itself may or may not be included.

        We will track any field on a custom object or standard entity. We will not track
        fields on Custom Settings, Platform Events, or Custom Metadata Types."""
        return (sobject_name.endswith("__c") or sobject_name.count("__") == 0) and (
            element.find("customSettingsType") is None if element else True
        )

    def _process_sfdx_release(self, zip_file, version):
        """Process an SFDX ZIP file for objects and fields"""
        for f in zip_file.namelist():
            path = PurePosixPath(f)
            if f.startswith("force-app/main/default/objects"):
                if path.suffixes == [".object-meta", ".xml"]:
                    sobject_name = path.name[: -len(".object-meta.xml")]
                    if sobject_name.count("__") == 1:
                        sobject_name = f"{version.package.namespace}{sobject_name}"

                    element = metadata_tree.fromstring(zip_file.read(f))

                    if self._should_process_object(
                        version.package.namespace, sobject_name, element
                    ):
                        self._process_object_element(sobject_name, element, version)
                    else:
                        # If this is an object type from which we shouldn't process any fields,
                        # track it in omit_sobjects so we can drop any fields later if we don't have
                        # the right information at time of processing.

                        # Note that the owning object may be in a dependency package, so we won't find it below.
                        if not self._should_process_object_fields(
                            sobject_name, element
                        ):
                            self.omit_sobjects.add(sobject_name)
                elif path.suffixes == [".field-meta", ".xml"]:
                    # To get the sObject name, we need to remove the `/fields/SomeField.field-meta.xml`
                    # and take the last path component

                    # Find the sObject metadata file
                    sobject_name = f"{path.parent.parent.stem}"
                    sobject_file = str(
                        path.parent.parent / f"{sobject_name}.object-meta.xml"
                    )
                    if sobject_name.count("__") == 1:
                        sobject_name = f"{version.package.namespace}{sobject_name}"

                    # If the object-meta file is locatable, load it so we can check
                    # if this is a Custom Setting.
                    if sobject_file in zip_file.namelist():
                        object_entity = metadata_tree.fromstring(
                            zip_file.read(sobject_file)
                        )
                    else:
                        object_entity = None

                    if self._should_process_object_fields(sobject_name, object_entity):
                        self._process_field_element(
                            sobject_name,
                            metadata_tree.fromstring(zip_file.read(f)),
                            version,
                        )

    def _process_object_element(self, sobject_name, element, version):
        """Process a <CustomObject> metadata entity, whether SFDX or MDAPI"""
        description_elem = getattr(element, "description", None)

        # SFDX release can check this before calling us, but MDAPI can't.
        if self._should_process_object(
            version.package.namespace, sobject_name, element
        ):
            self.sobjects[sobject_name].append(
                SObjectDetail(
                    version,
                    sobject_name,
                    element.label.text,
                    description_elem.text if description_elem is not None else "",
                )
            )

        # For MDAPI-format elements. No-op on SFDX.
        if self._should_process_object_fields(sobject_name, element):
            for field in element.findall("fields"):
                self._process_field_element(sobject_name, field, version)
        else:
            # If this is an object type from which we shouldn't process any fields,
            # track it in omit_sobjects so we can drop any fields later if we don't have
            # the right information at time of processing.

            # Note that the owning object may be in a dependency package, so we won't find it otherwise.
            self.omit_sobjects.add(sobject_name)

    def _process_field_element(self, sobject, field, version):
        """Process a field entity, which can be either a <fields> element
        in MDAPI format or a <CustomField> in SFDX"""
        # `element` may be either a `fields` element (in MDAPI)
        # or a `CustomField` (SFDX)
        # If this is a custom field, register its presence in this version
        field_name = f"{version.package.namespace}{field.fullName.text}"
        # get field help text value
        help_text_elem = field.find("inlineHelpText")
        # get field description text value
        description_text_elem = field.find("description")

        if "__" in field_name:
            field_type = field.type.text
            valid_values = ""

            length = ""

            if not field.find("formula"):
                if field_type in ["Text", "LongTextArea"]:
                    length = f" ({field.length.text})"
                elif field_type == "Number":
                    length = f" ({int(field.precision.text) - int(field.scale.text)}.{field.scale.text})"

            if field_type in ("Picklist", "MultiselectPicklist"):
                # There's two different ways of storing picklist values
                # (exclusive of Global Value Sets).
                # <picklist> is used prior to API 38.0: https://developer.salesforce.com/docs/atlas.en-us.api_meta.meta/api_meta/meta_picklist.htm
                # <valueSet> is used thereafter: https://developer.salesforce.com/docs/atlas.en-us.api_meta.meta/api_meta/meta_field_types.htm#meta_type_valueset
                if field.find("valueSet") is not None:
                    # Determine if this field uses a Global Value Set.
                    value_set = field.valueSet
                    if value_set.find("valueSetName") is not None:
                        value_set_name = value_set.find("valueSetName").text
                        valid_values = f"Global Value Set {value_set_name}"
                    else:
                        valueSetDefinition = value_set.valueSetDefinition
                        labels = [
                            value.find("label").text
                            if value.find("label")
                            else value.fullName.text
                            for value in valueSetDefinition.findall("value")
                        ]

                        valid_values = "; ".join(labels)
                elif field.find("picklist") is not None:
                    picklist = field.picklist
                    names = [
                        value.fullName.text
                        for value in picklist.findall("picklistValues")
                    ]

                    valid_values = "; ".join(names)
            elif field_type == "Lookup":
                target_sobject = field.referenceTo.text
                if target_sobject.count("__") == 1:
                    target_sobject = f"{version.package.namespace}{target_sobject}"
                field_type = f"Lookup to {target_sobject}"
                # Note: polymorphic custom fields are not allowed.
            elif field_type == "MasterDetail":
                target_sobject = field.referenceTo.text
                if target_sobject.count("__") == 1:
                    target_sobject = f"{version.package.namespace}{target_sobject}"
                field_type = f"Master-Detail Relationship to {target_sobject}"
                # Note: polymorphic custom fields are not allowed.

            fd = FieldDetail(
                version,
                sobject,
                field_name,
                field.label.text,
                f"{field_type}{length}",
                help_text_elem.text if help_text_elem else "",
                description_text_elem.text if description_text_elem else "",
                valid_values,
            )
            fully_qualified_name = f"{sobject}.{fd.api_name}"
            self.fields[fully_qualified_name].append(fd)

    def _write_object_results(self, file_handle):
        """Write to the given handle an output CSV containing the data dictionary for sObjects."""
        writer = csv.writer(file_handle)

        writer.writerow(
            [
                "Object Label",
                "Object API Name",
                "Object Description",
                "Version Introduced",
                "Version Deleted",
            ]
        )

        for sobject_name, versions in self.sobjects.items():
            # object_version.version.version yields the StrictVersion of the package version of this object version.
            versions = sorted(
                versions,
                key=lambda object_version: object_version.version.version,
                reverse=True,
            )
            first_version = versions[-1]
            last_version = versions[0]

            # Locate the version, if any, where this object was deleted.
            package_versions = sorted(
                self.package_versions[last_version.version.package]
            )
            if last_version.version.version != package_versions[-1]:
                deleted_version = package_versions[
                    package_versions.index(last_version.version.version) + 1
                ]
            else:
                deleted_version = None

            if sobject_name.endswith("__c"):
                writer.writerow(
                    [
                        last_version.label,
                        sobject_name,
                        last_version.description,
                        f"{first_version.version.package.package_name} {first_version.version.version}",
                        ""
                        if deleted_version is None
                        else f"{first_version.version.package.package_name} {deleted_version}",
                    ]
                )

    def _write_field_results(self, file_handle):
        """Write to the given handle an output CSV containing the data dictionary for fields."""
        writer = csv.writer(file_handle)

        writer.writerow(
            [
                "Object Label",
                "Object API Name",
                "Field Label",
                "Field API Name",
                "Type",
                "Picklist Values",
                "Help Text",
                "Field Description",
                "Version Introduced",
                "Version Picklist Values Last Changed",
                "Version Help Text Last Changed",
                "Version Deleted",
            ]
        )

        for field_name, field_versions in self.fields.items():
            # field_version.version.version yields the StrictVersion of the package version for this field version.
            versions = sorted(
                field_versions,
                key=lambda field_version: field_version.version.version,
                reverse=True,
            )
            first_version = versions[-1]
            last_version = versions[0]

            if last_version.sobject in self.omit_sobjects:
                continue

            # Locate the last versions where the valid values and the help text changed.
            valid_values_version = None
            for (index, version) in enumerate(versions[1:]):
                if version.valid_values != last_version.valid_values:
                    valid_values_version = versions[index]
                    break

            help_text_version = None
            for (index, version) in enumerate(versions[1:]):
                if version.help_text != last_version.help_text:
                    help_text_version = versions[index]
                    break

            # Locate the version, if any, where this field was deleted.
            package_versions = sorted(
                self.package_versions[last_version.version.package]
            )
            if last_version.version.version != package_versions[-1]:
                deleted_version = package_versions[
                    package_versions.index(last_version.version.version) + 1
                ]
            else:
                deleted_version = None

            # Find the sObject name, if possible, for this field.
            if last_version.sobject in self.sobjects:
                sobject_label = sorted(
                    self.sobjects[last_version.sobject],
                    key=lambda ver: ver.version.version,
                    reverse=True,
                )[0].label
            else:
                sobject_label = last_version.sobject

            writer.writerow(
                [
                    sobject_label,
                    last_version.sobject,
                    last_version.label,
                    last_version.api_name,
                    last_version.type,
                    last_version.valid_values,
                    last_version.help_text,
                    last_version.description,
                    f"{first_version.version.package.package_name} {first_version.version.version}",
                    f"{first_version.version.package.package_name} {valid_values_version.version.version}"
                    if valid_values_version
                    else "",
                    f"{first_version.version.package.package_name} {help_text_version.version.version}"
                    if help_text_version
                    else "",
                    ""
                    if deleted_version is None
                    else f"{first_version.version.package.package_name} {deleted_version}",
                ]
            )

    def _write_results(self):
        """Write the stored schema details to our destination CSVs"""
        with open(self.options["object_path"], "w") as object_file:
            self._write_object_results(object_file)

        with open(self.options["field_path"], "w") as field_file:
            self._write_field_results(field_file)

    def _version_from_tag_name(self, tag_name, prefix_release):
        """Parse a release's tag and return a StrictVersion"""
        return StrictVersion(tag_name[len(prefix_release) :])



from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.utils import process_bool_arg
from cumulusci.salesforce_api.exceptions import MetadataApiError
from cumulusci.salesforce_api.package_zip import InstallPackageZipBuilder
from cumulusci.tasks.salesforce import Deploy


class InstallPackageVersion(Deploy):
    task_options = {
        "name": {
            "description": "The name of the package to install.  Defaults to project__package__name_managed",
            "required": False,
        },
        "namespace": {
            "description": "The namespace of the package to install.  Defaults to project__package__namespace",
            "required": True,
        },
        "version": {
            "description": 'The version of the package to install.  "latest" and "latest_beta" can be used to trigger lookup via Github Releases on the repository.',
            "required": True,
        },
        "activateRSS": {
            "description": "If True, preserve the isActive state of "
            "Remote Site Settings and Content Security Policy "
            "in the package. Default: False."
        },
        "password": {"description": "The package password. Optional."},
        "retries": {"description": "Number of retries (default=5)"},
        "retry_interval": {
            "description": "Number of seconds to wait before the next retry (default=5),"
        },
        "retry_interval_add": {
            "description": "Number of seconds to add before each retry (default=30),"
        },
        "security_type": {
            "description": "Which users to install package for (FULL = all users, NONE = admins only)"
        },
    }

    def _init_options(self, kwargs):
        super(InstallPackageVersion, self)._init_options(kwargs)
        if "namespace" not in self.options:
            self.options["namespace"] = self.project_config.project__package__namespace
        if "name" not in self.options:
            self.options["name"] = (
                self.project_config.project__package__name_managed
                or self.project_config.project__package__name
                or self.options["namespace"]
            )
        if "retries" not in self.options:
            self.options["retries"] = 5
        if "retry_interval" not in self.options:
            self.options["retry_interval"] = 5
        if "retry_interval_add" not in self.options:
            self.options["retry_interval_add"] = 30
        version = self.options.get("version")
        if version == "latest":
            self.options["version"] = self.project_config.get_latest_version()
        elif version == "latest_beta":
            self.options["version"] = self.project_config.get_latest_version(beta=True)
        elif version == "previous":
            self.options["version"] = self.project_config.get_previous_version()
        self.options["activateRSS"] = process_bool_arg(self.options.get("activateRSS"))
        self.options["security_type"] = self.options.get("security_type", "FULL")
        if self.options["security_type"] not in ("FULL", "NONE", "PUSH"):
            raise TaskOptionsError(
                f"Unsupported value for security_type: {self.options['security_type']}"
            )

    def _get_api(self, path=None):
        package_zip = InstallPackageZipBuilder(
            namespace=self.options["namespace"],
            version=self.options["version"],
            activateRSS=self.options["activateRSS"],
            password=self.options.get("password"),
            securityType=self.options.get("security_type", "FULL"),
        )
        return self.api_class(self, package_zip(), purge_on_delete=False)

    def _run_task(self):
        self.logger.info(
            f"Installing {self.options['name']} release: {self.options['version']}"
        )
        self._retry()
        self.org_config.reset_installed_packages()

    def _try(self):
        api = self._get_api()
        api()

    def _is_retry_valid(self, e):
        if isinstance(e, MetadataApiError) and (
            "This package is not yet available" in str(e)
            or "InstalledPackage version number" in str(e)
        ):
            return True

    def freeze(self, step):
        options = self.options.copy()
        options["version"] = str(options["version"])
        name = options.pop("name")
        task_config = {"options": options, "checks": self.task_config.checks or []}
        ui_step = {
            "name": "Install {} {}".format(name, options["version"]),
            "kind": "managed",
            "is_required": True,
        }
        ui_step.update(step.task_config.get("ui_options", {}))
        ui_step.update(
            {
                "path": step.path,
                "step_num": str(step.step_num),
                "task_class": self.task_config.class_path,
                "task_config": task_config,
                "source": step.project_config.source.frozenspec,
            }
        )
        return [ui_step]



from unittest.mock import Mock
import unittest

from cumulusci.tasks.salesforce.license_preflights import (
    GetAvailableLicenses,
    GetAvailablePermissionSetLicenses,
)
from .util import create_task


class TestLicensePreflights(unittest.TestCase):
    def test_license_preflight(self):
        task = create_task(GetAvailableLicenses, {})
        task._init_api = Mock()
        task._init_api.return_value.query.return_value = {
            "totalSize": 2,
            "records": [
                {"LicenseDefinitionKey": "TEST1"},
                {"LicenseDefinitionKey": "TEST2"},
            ],
        }
        task()

        task._init_api.return_value.query.assert_called_once_with(
            "SELECT LicenseDefinitionKey FROM UserLicense"
        )
        assert task.return_values == ["TEST1", "TEST2"]

    def test_psl_preflight(self):
        task = create_task(GetAvailablePermissionSetLicenses, {})
        task._init_api = Mock()

        task._init_api.return_value.query.return_value = {
            "totalSize": 2,
            "records": [
                {"PermissionSetLicenseKey": "TEST1"},
                {"PermissionSetLicenseKey": "TEST2"},
            ],
        }
        task()

        task._init_api.return_value.query.assert_called_once_with(
            "SELECT PermissionSetLicenseKey FROM PermissionSetLicense"
        )
        assert task.return_values == ["TEST1", "TEST2"]



import base64
import io
from unittest import mock
import unittest
import zipfile

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.salesforce_api.exceptions import MetadataApiError
from cumulusci.tasks.salesforce import InstallPackageVersion
from cumulusci.tests.util import create_project_config
from .util import create_task


class TestInstallPackageVersion(unittest.TestCase):
    def test_run_task_with_retry(self):
        project_config = create_project_config()
        project_config.get_latest_version = mock.Mock(return_value="1.0")
        project_config.config["project"]["package"]["namespace"] = "ns"
        task = create_task(InstallPackageVersion, {"version": "latest"}, project_config)
        not_yet = MetadataApiError("This package is not yet available", None)
        api = mock.Mock(side_effect=[not_yet, None])
        task.api_class = mock.Mock(return_value=api)
        task()
        self.assertEqual(2, api.call_count)

    def test_run_task__options(self):
        project_config = create_project_config()
        project_config.get_latest_version = mock.Mock(return_value="1.0 (Beta 1)")
        project_config.config["project"]["package"]["namespace"] = "ns"
        task = create_task(
            InstallPackageVersion,
            {
                "version": "latest_beta",
                "activateRSS": True,
                "password": "astro",
                "security_type": "NONE",
            },
            project_config,
        )
        api = task._get_api()
        zf = zipfile.ZipFile(io.BytesIO(base64.b64decode(api.package_zip)), "r")
        package_xml = zf.read("installedPackages/ns.installedPackage")
        self.assertIn(b"<activateRSS>true</activateRSS", package_xml)
        self.assertIn(b"<password>astro</password>", package_xml)
        self.assertIn(b"<securityType>NONE</securityType>", package_xml)

    def test_run_task__bad_security_type(self):
        project_config = create_project_config()
        project_config.get_latest_version = mock.Mock(return_value="1.0")
        project_config.config["project"]["package"]["namespace"] = "ns"
        with self.assertRaises(TaskOptionsError):
            create_task(
                InstallPackageVersion,
                {"version": "latest", "security_type": "BOGUS"},
                project_config,
            )



from cumulusci.core.github import get_github_api_for_repo
from cumulusci.core.tasks import BaseTask


class BaseGithubTask(BaseTask):
    def _init_task(self):
        self.github_config = self.project_config.keychain.get_service("github")
        self.github = get_github_api_for_repo(
            self.project_config.keychain,
            self.project_config.repo_owner,
            self.project_config.repo_name,
        )

    def get_repo(self):
        return self.github.repository(
            self.project_config.repo_owner, self.project_config.repo_name
        )



import json
import unittest
from unittest import mock
import responses
import yaml

from cumulusci.tasks.bulkdata import GenerateMapping
from cumulusci.utils import temporary_dir
from cumulusci.tasks.bulkdata.tests.utils import _make_task


class TestMappingGenerator(unittest.TestCase):
    def test_defaults_options(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        self.assertEqual([], t.options["ignore"])
        self.assertEqual("", t.options["namespace_prefix"])

    def test_postfixes_underscores_to_namespace(self):
        t = _make_task(
            GenerateMapping, {"options": {"namespace_prefix": "t", "path": "t"}}
        )

        self.assertEqual("t__", t.options["namespace_prefix"])

    def test_splits_ignore_string(self):
        t = _make_task(
            GenerateMapping, {"options": {"ignore": "Account, Contact", "path": "t"}}
        )

        self.assertEqual(["Account", "Contact"], t.options["ignore"])

    def test_accepts_ignore_list(self):
        t = _make_task(
            GenerateMapping,
            {"options": {"ignore": ["Account", "Contact"], "path": "t"}},
        )

        self.assertEqual(["Account", "Contact"], t.options["ignore"])

    def test_is_any_custom_api_name(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        self.assertTrue(t._is_any_custom_api_name("Custom__c"))
        self.assertFalse(t._is_any_custom_api_name("Standard"))

    def test_is_our_custom_api_name(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        self.assertTrue(t._is_our_custom_api_name("Custom__c"))
        self.assertFalse(t._is_our_custom_api_name("Standard"))
        self.assertFalse(t._is_our_custom_api_name("t__Custom__c"))
        self.assertFalse(t._is_our_custom_api_name("f__Custom__c"))

        t.options["namespace_prefix"] = "t__"
        self.assertTrue(t._is_our_custom_api_name("Custom__c"))
        self.assertTrue(t._is_our_custom_api_name("t__Custom__c"))
        self.assertFalse(t._is_our_custom_api_name("f__Custom__c"))

    def test_is_core_field(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        self.assertTrue(t._is_core_field("Id"))
        self.assertFalse(t._is_core_field("Custom__c"))

    def test_is_object_mappable(self):
        t = _make_task(GenerateMapping, {"options": {"ignore": "Account", "path": "t"}})

        self.assertTrue(
            t._is_object_mappable({"name": "Contact", "customSetting": False})
        )
        self.assertFalse(
            t._is_object_mappable({"name": "Account", "customSetting": False})
        )
        self.assertFalse(
            t._is_object_mappable(
                {"name": "Contact__ChangeEvent", "customSetting": False}
            )
        )
        self.assertFalse(
            t._is_object_mappable({"name": "Custom__c", "customSetting": True})
        )

    def test_is_field_mappable(self):
        t = _make_task(
            GenerateMapping, {"options": {"ignore": "Account.ParentId", "path": "t"}}
        )

        t.mapping_objects = ["Account", "Contact"]

        self.assertTrue(
            t._is_field_mappable(
                "Account",
                {"name": "Name", "type": "string", "label": "Name", "createable": True},
            )
        )
        self.assertFalse(
            t._is_field_mappable(
                "Account",
                {"name": "Name", "type": "base64", "label": "Name", "createable": True},
            )
        )
        self.assertFalse(
            t._is_field_mappable(
                "Account",
                {
                    "name": "Name",
                    "type": "string",
                    "label": "Name (Deprecated)",
                    "createable": True,
                },
            )
        )
        self.assertFalse(
            t._is_field_mappable(
                "Account",
                {
                    "name": "ParentId",
                    "type": "reference",
                    "label": "Parent",
                    "createable": True,
                    "referenceTo": ["Account"],
                },
            )
        )
        self.assertFalse(
            t._is_field_mappable(
                "Account",
                {
                    "name": "Name",
                    "type": "string",
                    "label": "Name",
                    "createable": False,
                },
            )
        )
        self.assertFalse(
            t._is_field_mappable(
                "Contact",
                {
                    "name": "OwnerId",
                    "type": "reference",
                    "label": "Owner",
                    "createable": True,
                    "referenceTo": ["User", "Group"],
                },
            )
        )

    def test_has_our_custom_fields(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        self.assertTrue(t._has_our_custom_fields({"fields": [{"name": "Custom__c"}]}))
        self.assertTrue(
            t._has_our_custom_fields(
                {"fields": [{"name": "Custom__c"}, {"name": "Standard"}]}
            )
        )
        self.assertFalse(t._has_our_custom_fields({"fields": [{"name": "Standard"}]}))
        self.assertFalse(t._has_our_custom_fields({"fields": []}))

    def test_is_lookup_to_included_object(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        t.mapping_objects = ["Account"]

        self.assertTrue(
            t._is_lookup_to_included_object(
                {"type": "reference", "referenceTo": ["Account"]}
            )
        )
        self.assertFalse(
            t._is_lookup_to_included_object(
                {"type": "reference", "referenceTo": ["Contact"]}
            )
        )
        self.assertFalse(
            t._is_lookup_to_included_object(
                {"type": "reference", "referenceTo": ["Account", "Contact"]}
            )
        )

    def _prepare_describe_mock(self, task, describe_data):
        responses.add(
            method="GET",
            url=f"{task.org_config.instance_url}/services/data/v45.0/sobjects",
            body=json.dumps(
                {
                    "sobjects": [
                        {"name": s, "customSetting": False} for s in describe_data
                    ]
                }
            ),
            status=200,
        )
        for s in describe_data:
            body = {"name": s, "customSetting": False}
            body.update(describe_data[s])
            responses.add(
                method="GET",
                url=f"{task.org_config.instance_url}/services/data/v45.0/sobjects/{s}/describe",
                body=json.dumps(body),
                status=200,
            )

    def _mock_field(self, name, field_type="string", **kwargs):
        field_data = {
            "name": name,
            "type": field_type,
            "createable": True,
            "nillable": True,
            "label": name,
        }
        field_data.update(kwargs)
        return field_data

    @responses.activate
    def test_run_task(self):
        t = _make_task(GenerateMapping, {"options": {"path": "mapping.yaml"}})
        t.project_config.project__package__api_version = "45.0"
        describe_data = {
            "Parent": {
                "fields": [self._mock_field("Id"), self._mock_field("Custom__c")]
            },
            "Child__c": {
                "fields": [
                    self._mock_field("Id"),
                    self._mock_field(
                        "Account__c",
                        field_type="reference",
                        referenceTo=["Parent"],
                        relationshipOrder=None,
                    ),
                ]
            },
        }

        self._prepare_describe_mock(t, describe_data)
        with temporary_dir():
            t()

            with open("mapping.yaml", "r") as fh:
                content = yaml.safe_load(fh)

            self.assertEqual(["Insert Parent", "Insert Child__c"], list(content.keys()))
            self.assertEqual("Parent", t.mapping["Insert Parent"]["sf_object"])
            self.assertEqual("Parent", t.mapping["Insert Parent"]["table"])
            self.assertEqual(
                ["Id", "Custom__c"], list(t.mapping["Insert Parent"]["fields"].keys())
            )
            self.assertEqual("sf_id", t.mapping["Insert Parent"]["fields"]["Id"])
            self.assertEqual(
                "Custom__c", t.mapping["Insert Parent"]["fields"]["Custom__c"]
            )

            self.assertEqual("Child__c", t.mapping["Insert Child__c"]["sf_object"])
            self.assertEqual("Child__c", t.mapping["Insert Child__c"]["table"])
            self.assertEqual(
                ["Id"], list(t.mapping["Insert Child__c"]["fields"].keys())
            )
            self.assertEqual(
                ["Account__c"], list(t.mapping["Insert Child__c"]["lookups"].keys())
            )
            self.assertEqual("sf_id", t.mapping["Insert Child__c"]["fields"]["Id"])
            self.assertEqual(
                "Parent", t.mapping["Insert Child__c"]["lookups"]["Account__c"]["table"]
            )

    @responses.activate
    def test_collect_objects__simple_custom_objects(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        t.project_config.project__package__api_version = "45.0"

        describe_data = {
            "Account": {
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")]
            },
            "Contact": {"fields": [self._mock_field("Name")]},
            "Custom__c": {
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")]
            },
        }

        self._prepare_describe_mock(t, describe_data)
        t._init_task()
        t._collect_objects()

        self.assertEqual(set(["Account", "Custom__c"]), set(t.mapping_objects))

    @responses.activate
    def test_collect_objects__custom_lookup_fields(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        t.project_config.project__package__api_version = "45.0"

        describe_data = {
            "Account": {
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")]
            },
            "Contact": {"fields": [self._mock_field("Name")]},
            "Custom__c": {
                "fields": [
                    self._mock_field("Name"),
                    self._mock_field("Custom__c"),
                    self._mock_field(
                        "Lookup__c",
                        field_type="reference",
                        relationshipOrder=None,
                        referenceTo=["Contact"],
                    ),
                ]
            },
        }

        self._prepare_describe_mock(t, describe_data)
        t._init_task()
        t._collect_objects()

        self.assertEqual(
            set(["Account", "Custom__c", "Contact"]), set(t.mapping_objects)
        )

    @responses.activate
    def test_collect_objects__master_detail_fields(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        t.project_config.project__package__api_version = "45.0"

        describe_data = {
            "Account": {
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")]
            },
            "Opportunity": {"fields": [self._mock_field("Name")]},
            "OpportunityLineItem": {
                "fields": [
                    self._mock_field("Name"),
                    self._mock_field("Custom__c"),
                    self._mock_field(
                        "OpportunityId",
                        field_type="reference",
                        relationshipOrder=1,
                        referenceTo=["Opportunity"],
                    ),
                ]
            },
        }

        self._prepare_describe_mock(t, describe_data)
        t._init_task()
        t._collect_objects()

        self.assertEqual(
            set(["Account", "OpportunityLineItem", "Opportunity"]),
            set(t.mapping_objects),
        )

    @responses.activate
    def test_collect_objects__duplicate_references(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        t.project_config.project__package__api_version = "45.0"

        describe_data = {
            "Account": {
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")]
            },
            "Opportunity": {"fields": [self._mock_field("Name")]},
            "OpportunityLineItem": {
                "fields": [
                    self._mock_field("Name"),
                    self._mock_field("Custom__c"),
                    self._mock_field(
                        "OpportunityId",
                        field_type="reference",
                        relationshipOrder=1,
                        referenceTo=["Opportunity"],
                    ),
                    self._mock_field(
                        "CustomLookup__c",
                        field_type="reference",
                        relationshipOrder=None,
                        referenceTo=["Opportunity"],
                    ),
                ]
            },
        }

        self._prepare_describe_mock(t, describe_data)
        t._init_task()
        t._collect_objects()

        self.assertEqual(
            set(["Account", "OpportunityLineItem", "Opportunity"]),
            set(t.mapping_objects),
        )

    def test_build_schema(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        t.mapping_objects = ["Account", "Opportunity", "Child__c"]
        stage_name = self._mock_field("StageName")
        stage_name["nillable"] = False
        t.describes = {
            "Account": {
                "fields": [self._mock_field("Name"), self._mock_field("Industry")]
            },
            "Opportunity": {"fields": [self._mock_field("Name"), stage_name]},
            "Child__c": {
                "fields": [
                    self._mock_field("Name"),
                    self._mock_field("Test__c"),
                    self._mock_field("Attachment__c", field_type="base64"),
                ]
            },
        }

        t._build_schema()
        self.assertEqual(
            {
                "Account": {"Name": self._mock_field("Name")},
                "Opportunity": {
                    "Name": self._mock_field("Name"),
                    "StageName": stage_name,
                },
                "Child__c": {
                    "Name": self._mock_field("Name"),
                    "Test__c": self._mock_field("Test__c"),
                },
            },
            t.schema,
        )

    def test_build_schema__tracks_references(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        t.mapping_objects = ["Account", "Opportunity"]
        t.describes = {
            "Account": {"fields": [self._mock_field("Name")]},
            "Opportunity": {
                "fields": [
                    self._mock_field("Name"),
                    self._mock_field(
                        "AccountId",
                        field_type="reference",
                        referenceTo=["Account"],
                        relationshipOrder=1,
                    ),
                ]
            },
        }

        t._build_schema()
        self.assertEqual({"Opportunity": {"Account": set(["AccountId"])}}, t.refs)

    def test_build_schema__includes_recordtypeid(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        t.mapping_objects = ["Account", "Opportunity"]
        t.describes = {
            "Account": {"fields": [self._mock_field("Name")]},
            "Opportunity": {
                "fields": [
                    self._mock_field("Name"),
                    self._mock_field(
                        "AccountId",
                        field_type="reference",
                        referenceTo=["Account"],
                        relationshipOrder=1,
                    ),
                    self._mock_field("RecordTypeId"),
                ],
                "recordTypeInfos": [{"Name": "Master"}, {"Name": "Donation"}],
            },
        }

        t._build_schema()
        self.assertIn("RecordTypeId", t.schema["Opportunity"])
        self.assertNotIn("RecordTypeId", t.schema["Account"])

    @mock.patch("click.prompt")
    def test_build_mapping(self, prompt):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        prompt.return_value = "Account"

        t.schema = {
            "Account": {
                "Id": self._mock_field("Id"),
                "Name": self._mock_field("Name"),
                "Dependent__c": self._mock_field(
                    "Dependent__c", field_type="reference", referenceTo=["Child__c"]
                ),
            },
            "Child__c": {
                "Id": self._mock_field("Id"),
                "Name": self._mock_field("Name"),
                "Account__c": self._mock_field(
                    "Account__c", field_type="reference", referenceTo=["Account"]
                ),
                "Self__c": self._mock_field(
                    "Self__c", field_type="reference", referenceTo=["Child__c"]
                ),
            },
        }
        t.refs = {
            "Child__c": {"Account": set(["Account__c"])},
            "Account": {"Child__c": set(["Dependent__c"])},
        }

        t._build_mapping()
        self.assertEqual(["Insert Account", "Insert Child__c"], list(t.mapping.keys()))
        self.assertEqual("Account", t.mapping["Insert Account"]["sf_object"])
        self.assertEqual("Account", t.mapping["Insert Account"]["table"])
        self.assertEqual(
            ["Id", "Name"], list(t.mapping["Insert Account"]["fields"].keys())
        )
        self.assertEqual("sf_id", t.mapping["Insert Account"]["fields"]["Id"])
        self.assertEqual("Name", t.mapping["Insert Account"]["fields"]["Name"])
        self.assertEqual(
            ["Dependent__c"], list(t.mapping["Insert Account"]["lookups"].keys())
        )
        self.assertEqual(
            "Child__c", t.mapping["Insert Account"]["lookups"]["Dependent__c"]["table"]
        )

        self.assertEqual("Child__c", t.mapping["Insert Child__c"]["sf_object"])
        self.assertEqual("Child__c", t.mapping["Insert Child__c"]["table"])
        self.assertEqual(
            ["Id", "Name"], list(t.mapping["Insert Child__c"]["fields"].keys())
        )
        self.assertEqual(
            ["Account__c", "Self__c"],
            list(t.mapping["Insert Child__c"]["lookups"].keys()),
        )
        self.assertEqual("sf_id", t.mapping["Insert Child__c"]["fields"]["Id"])
        self.assertEqual("Name", t.mapping["Insert Child__c"]["fields"]["Name"])
        self.assertEqual(
            "Account", t.mapping["Insert Child__c"]["lookups"]["Account__c"]["table"]
        )
        self.assertEqual(
            "Child__c", t.mapping["Insert Child__c"]["lookups"]["Self__c"]["table"]
        )

    def test_build_mapping__warns_polymorphic_lookups(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        t.mapping_objects = ["Account", "Contact", "Custom__c"]
        t.schema = {
            "Account": {"Name": self._mock_field("Name")},
            "Contact": {"Name": self._mock_field("Name")},
            "Custom__c": {
                "Name": self._mock_field("Name"),
                "PolyLookup__c": self._mock_field(
                    "PolyLookup__c",
                    field_type="reference",
                    referenceTo=["Account", "Contact"],
                ),
            },
        }
        t.refs = {
            "Custom__c": {
                "Account": set(["PolyLookup__c"]),
                "Contact": set(["PolyLookup__c"]),
            }
        }
        t.logger = mock.Mock()

        t._build_mapping()
        t.logger.warning.assert_called_once_with(
            "Field Custom__c.PolyLookup__c is a polymorphic lookup, which is not supported"
        )

    def test_split_dependencies__no_cycles(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        stack = t._split_dependencies(
            set(["Account", "Contact", "Opportunity", "Custom__c"]),
            {
                "Contact": {"Account": set(["AccountId"])},
                "Opportunity": {
                    "Account": set(["AccountId"]),
                    "Contact": set(["Primary_Contact__c"]),
                },
                "Custom__c": {
                    "Account": set(["Account__c"]),
                    "Contact": set(["Contact__c"]),
                    "Opportunity": set(["Opp__c"]),
                },
            },
        )

        self.assertEqual(["Account", "Contact", "Opportunity", "Custom__c"], stack)

    @mock.patch("click.prompt")
    def test_split_dependencies__interviews_for_cycles(self, prompt):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        prompt.return_value = "Account"

        self.assertEqual(
            ["Custom__c", "Account", "Contact", "Opportunity"],
            t._split_dependencies(
                set(["Account", "Contact", "Opportunity", "Custom__c"]),
                {
                    "Account": {"Contact": set(["Primary_Contact__c"])},
                    "Contact": {"Account": set(["AccountId"])},
                    "Opportunity": {
                        "Account": set(["AccountId"]),
                        "Contact": set(["Primary_Contact__c"]),
                    },
                },
            ),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os.path
from setuptools import setup
from pkgutil import walk_packages


def find_packages(path=["."], prefix=""):
    yield prefix
    prefix = prefix + "."
    for _, name, ispkg in walk_packages(path, prefix):
        if ispkg:
            yield name


with open(os.path.join("cumulusci", "version.txt"), "r") as version_file:
    version = version_file.read().strip()

with open("README.rst", "rb") as readme_file:
    readme = readme_file.read().decode("utf-8")

with open("HISTORY.rst", "rb") as history_file:
    history = history_file.read().decode("utf-8")

with open("requirements.txt") as requirements_file:
    requirements = [
        req.split("#")[0].strip() for req in requirements_file.read().splitlines()
    ]

with open("requirements_dev.txt") as dev_requirements_file:
    test_requirements = [
        req
        for req in dev_requirements_file.read().splitlines()
        if not req.startswith("-")
    ]


setup(
    name="cumulusci",
    version=version,
    description="Build and release tools for Salesforce developers",
    long_description=readme + u"\n\n" + history,
    long_description_content_type="text/x-rst",
    author="Salesforce.org",
    author_email="jlantz@salesforce.com",
    url="https://github.com/SFDO-Tooling/CumulusCI",
    packages=list(find_packages(["cumulusci"], "cumulusci")),
    package_dir={"cumulusci": "cumulusci"},
    entry_points={"console_scripts": ["cci=cumulusci.cli.cci:main"]},
    include_package_data=True,
    install_requires=requirements,
    license="BSD license",
    zip_safe=False,
    keywords="cumulusci",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    test_suite="cumulusci.core.tests",
    tests_require=test_requirements,
    python_requires=">=3.6",
)



""" a task for waiting on a Batch Apex job to complete """
from datetime import datetime
from typing import Sequence, Optional

from cumulusci.utils import parse_api_datetime
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.core.exceptions import SalesforceException

COMPLETED_STATUSES = ["Completed", "Aborted", "Failed"]
STOPPED_STATUSES = ["Aborted"]
FAILED_STATUSES = ["Failed"]


class BatchApexWait(BaseSalesforceApiTask):
    """ BatchApexWait polls an org until the latest batch job
        for an apex class completes or fails."""

    name = "BatchApexWait"
    original_created_date = None

    task_options = {
        "class_name": {
            "description": "Name of the Apex class to wait for.",
            "required": True,
        },
        "poll_interval": {
            "description": "Seconds to wait before polling for batch job completion. "
            "Defaults to 10 seconds."
        },
    }

    def _run_task(self):
        self.poll_interval_s = int(self.options.get("poll_interval", 10))

        self._poll()  # will block until poll_complete

        self.logger.info("Job is complete.")

        summary = self.summarize_subjobs(self.subjobs)
        failed_batches = self.failed_batches(self.subjobs)
        job_aborted = summary["AnyAborted"]
        job_failed = summary[
            "AnyFailed"
        ]  # note that a failed sub-job is different than a failed batch

        # per https://help.salesforce.com/articleView?id=code_apex_job.htm&type=5
        if job_aborted:
            raise SalesforceException("Job was aborted by a user.")
        elif job_failed:
            raise SalesforceException("Job experienced a system failure.")
        elif failed_batches:
            self.logger.info("There have been some batch failures.")
            raise SalesforceException(
                f"There were batch errors: {repr(failed_batches)}"
            )
        elif not summary["CountsAddUp"]:
            self.logger.info("The final record counts do not add up.")
            self.logger.info("This is probably related to W-1132237")
            self.logger.info(repr(summary))

        if len(self.subjobs) > 1:
            subjob_summary = f" in {len(self.subjobs)} sub-jobs"
        else:
            subjob_summary = ""

        self.logger.info(
            f"{self.options['class_name']} took {summary['ElapsedTime']} seconds to process {summary['TotalJobItems']} batches{subjob_summary}."
        )

    def failed_batches(self, subjobs: Sequence[dict]):
        failed_batches = []
        for subjob in subjobs:
            if subjob["NumberOfErrors"]:
                failed_batches.append(
                    {
                        key: value
                        for key, value in subjob.items()
                        if key
                        in {
                            "Id",
                            "Status",
                            "ExtendedStatus",
                            "NumberOfErrors",
                            "JobItemsProcessed",
                            "TotalJobItems",
                        }
                    }
                )
        return failed_batches

    def _poll_action(self):
        # get batch status

        if not self.original_created_date:
            query_results = self.tooling.query(self._batch_query(date_limit=None))
            if not query_results["records"]:
                raise SalesforceException(f"No {self.options['class_name']} job found.")
            self.original_created_date = parse_api_datetime(
                query_results["records"][0]["CreatedDate"]
            )
        else:
            query_results = self.tooling.query(
                self._batch_query(date_limit=self.original_created_date)
            )

        self.subjobs = query_results["records"]
        current_subjob = self.subjobs[0]

        summary = self.summarize_subjobs(self.subjobs)

        if len(self.subjobs) > 1:
            subjob_info = f" in {len(self.subjobs)} sub-jobs."
        else:
            subjob_info = ""

        self.logger.info(
            f"{self.options['class_name']}: "
            f"Job: {current_subjob['Id']} "
            f"{summary['JobItemsProcessed']} of {summary['TotalJobItems']} "
            f"({summary['NumberOfErrors']} failures)" + subjob_info
        )

        self.poll_complete = summary["Completed"]

    def summarize_subjobs(self, subjobs: Sequence[dict]):
        def reduce_key(valname: str, summary_func):
            return summary_func(subjob[valname] for subjob in subjobs)

        rc = {
            "JobItemsProcessed": reduce_key("JobItemsProcessed", sum),
            "TotalJobItems": reduce_key("TotalJobItems", sum),
            "NumberOfErrors": reduce_key("NumberOfErrors", sum),
            "Completed": all(
                subjob["Status"] in COMPLETED_STATUSES for subjob in subjobs
            ),
            "AnyAborted": any(
                subjob["Status"] in STOPPED_STATUSES for subjob in subjobs
            ),
            "AnyFailed": any(subjob["Status"] in FAILED_STATUSES for subjob in subjobs),
        }
        rc["Success"] = rc["NumberOfErrors"] == 0
        rc["ElapsedTime"] = self.elapsed_time(subjobs)
        rc["CountsAddUp"] = rc["JobItemsProcessed"] == rc["TotalJobItems"]
        return rc

    def elapsed_time(self, subjobs: Sequence[dict]):
        """ returns the time (in seconds) that the subjobs took, if complete """
        completed_dates = [
            subjob["CompletedDate"] for subjob in subjobs if subjob.get("CompletedDate")
        ]
        if completed_dates:
            most_recently_completed = max(completed_dates)
            completed_date = parse_api_datetime(most_recently_completed)
        else:
            completed_date = datetime.now()
        created_date = parse_api_datetime(
            min(subjob["CreatedDate"] for subjob in subjobs)
        )
        td = completed_date - created_date
        return td.total_seconds()

    def _batch_query(self, date_limit: Optional[datetime] = None):
        if not date_limit:
            limit = " LIMIT 1 "
            date_clause = "  "
        else:
            limit = " "
            date_clause = f" AND CreatedDate >= {date_limit.isoformat()}Z "

        query = (
            "SELECT Id, ApexClass.Name, Status, ExtendedStatus, TotalJobItems, "
            "JobItemsProcessed, NumberOfErrors, CreatedDate, CompletedDate "
            "FROM AsyncApexJob "
            "WHERE JobType='BatchApex' "
            + f"AND ApexClass.Name='{self.options['class_name']}' "
            + date_clause
            + " ORDER BY CreatedDate DESC "
            + limit
        )
        return query



import pytest

from cumulusci.cli.logger import init_logger
from cumulusci.cli.runtime import CliRuntime
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection
from cumulusci.core.config import TaskConfig


def pytest_addoption(parser, pluginmanager):
    parser.addoption("--org", action="store", default=None, help="org to use")


@pytest.fixture(scope="session")
def runtime():
    """Get the CumulusCI runtime for the current working directory."""
    init_logger()
    return CliRuntime()


@pytest.fixture(scope="session")
def project_config(runtime):
    """Get the project config for the current working directory."""
    return runtime.project_config


@pytest.fixture(scope="session")
def org_config(request, runtime):
    """Get an org config with an active access token.

    Specify the org name using the --org option when running pytest.
    Or else it will use your default CCI org.
    """
    org_name = request.config.getoption("--org")
    org_name, org_config = runtime.get_org(org_name)
    org_config.refresh_oauth_token(runtime.keychain)
    return org_config


@pytest.fixture
def sf(request, project_config, org_config):
    """Get a simple-salesforce client for org_config."""
    sf = get_simple_salesforce_connection(project_config, org_config)
    return sf


@pytest.fixture
def create_task(request, project_config, org_config):
    """Get a task _factory_ which can be used to construct task instances.
    """
    session_project_config = project_config
    session_org_config = org_config

    def create_task(task_class, options=None, project_config=None, org_config=None):
        project_config = project_config or session_project_config
        org_config = org_config or session_org_config
        options = options or {}

        task_config = TaskConfig({"options": options})

        return task_class(project_config, task_config, org_config)

    return create_task



from collections import defaultdict

import click
import yaml

from cumulusci.core.utils import process_list_arg
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class GenerateMapping(BaseSalesforceApiTask):
    task_docs = """
    Generate a mapping file for use with the `extract_dataset` and `load_dataset` tasks.
    This task will examine the schema in the specified org and attempt to infer a
    mapping suitable for extracting data in packaged and custom objects as well as
    customized standard objects.

    Mappings must be serializable, and hence must resolve reference cycles - situations
    where Object A refers to B, and B also refers to A. Mapping generation will stop
    and request user input to resolve such cycles by identifying the correct load order.
    Alternately, specify the `ignore` option with the name of one of the
    lookup fields to suppress it and break the cycle. `ignore` can be specified as a list in
    `cumulusci.yml` or as a comma-separated string at the command line.

    In most cases, the mapping generated will need minor tweaking by the user. Note
    that the mapping omits features that are not currently well supported by the
    `extract_dataset` and `load_dataset` tasks, such as references to
    the `User` object.
    """

    task_options = {
        "path": {"description": "Location to write the mapping file", "required": True},
        "namespace_prefix": {"description": "The namespace prefix to use"},
        "ignore": {
            "description": "Object API names, or fields in Object.Field format, to ignore"
        },
    }

    core_fields = ["Id", "Name", "FirstName", "LastName"]

    def _init_options(self, kwargs):
        super(GenerateMapping, self)._init_options(kwargs)
        if "namespace_prefix" not in self.options:
            self.options["namespace_prefix"] = ""

        if self.options["namespace_prefix"] and not self.options[
            "namespace_prefix"
        ].endswith("__"):
            self.options["namespace_prefix"] += "__"

        self.options["ignore"] = process_list_arg(self.options.get("ignore", []))

    def _run_task(self):
        self.logger.info("Collecting sObject information")
        self._collect_objects()
        self._build_schema()
        filename = self.options["path"]
        self.logger.info(f"Creating mapping schema {filename}")
        self._build_mapping()
        with open(filename, "w") as f:
            yaml.dump(self.mapping, f, sort_keys=False)

    def _collect_objects(self):
        """Walk the global describe and identify the sObjects we need to include in a minimal operation."""
        self.mapping_objects = []

        # Cache the global describe, which we'll walk.
        self.global_describe = self.sf.describe()

        # First, we'll get a list of all objects that are either
        # (a) custom, no namespace
        # (b) custom, with our namespace
        # (c) not ours (standard or other package), but have fields with our namespace or no namespace
        self.describes = {}  # Cache per-object describes for efficiency
        for obj in self.global_describe["sobjects"]:
            self.describes[obj["name"]] = getattr(self.sf, obj["name"]).describe()
            if self._is_our_custom_api_name(obj["name"]) or self._has_our_custom_fields(
                self.describes[obj["name"]]
            ):
                if self._is_object_mappable(obj):
                    self.mapping_objects.append(obj["name"])

        # Add any objects that are required by our own,
        # meaning any object we are looking up to with a custom field,
        # or any master-detail parent of any included object.
        index = 0
        while index < len(self.mapping_objects):
            obj = self.mapping_objects[index]
            for field in self.describes[obj]["fields"]:
                if field["type"] == "reference":
                    if field["relationshipOrder"] == 1 or self._is_any_custom_api_name(
                        field["name"]
                    ):
                        self.mapping_objects.extend(
                            [
                                obj
                                for obj in field["referenceTo"]
                                if obj not in self.mapping_objects
                                and self._is_object_mappable(self.describes[obj])
                            ]
                        )

            index += 1

    def _build_schema(self):
        """Convert self.mapping_objects into a schema, including field details and interobject references,
        in self.schema and self.refs"""

        # Now, find all the fields we need to include.
        # For custom objects, we include all custom fields. This includes custom objects
        # that our package doesn't own.
        # For standard objects, we include all custom fields, all required standard fields,
        # and master-detail relationships. Required means createable and not nillable.
        # In all cases, ensure that RecordTypeId is included if and only if there are Record Types
        self.schema = {}
        self.refs = defaultdict(lambda: defaultdict(set))
        for obj in self.mapping_objects:
            self.schema[obj] = {}

            for field in self.describes[obj]["fields"]:
                if any(
                    [
                        self._is_any_custom_api_name(field["name"]),
                        self._is_core_field(field["name"]),
                        self._is_required_field(field),
                        self._is_lookup_to_included_object(field),
                    ]
                ):
                    if self._is_field_mappable(obj, field):
                        self.schema[obj][field["name"]] = field

                        if field["type"] == "reference":
                            for target in field["referenceTo"]:
                                # We've already vetted that this field is referencing
                                # included objects, via `_is_field_mappable()`
                                if target != obj:
                                    self.refs[obj][target].add(field["name"])
                if (
                    field["name"] == "RecordTypeId"
                    and len(self.describes[obj]["recordTypeInfos"]) > 1
                ):
                    # "Master" is included even if no RTs.
                    self.schema[obj][field["name"]] = field

    def _build_mapping(self):
        """Output self.schema in mapping file format by constructing a dict and serializing to YAML"""
        objs = set(self.schema.keys())
        stack = self._split_dependencies(objs, self.refs)

        field_sort = (
            lambda f: "  " + f
            if f == "Id"
            else (" " + f if f in self.core_fields else f)
        )

        self.mapping = {}
        for obj in stack:
            key = f"Insert {obj}"
            self.mapping[key] = {}
            self.mapping[key]["sf_object"] = f"{obj}"
            self.mapping[key]["table"] = f"{obj}"
            fields = []
            lookups = []
            for field in self.schema[obj].values():
                if field["type"] == "reference" and field["name"] != "RecordTypeId":
                    lookups.append(field["name"])
                else:
                    fields.append(field["name"])
            self.mapping[key]["fields"] = {}
            if fields:
                if "Id" not in fields:
                    fields.append("Id")
                fields.sort(key=field_sort)
                for field in fields:
                    self.mapping[key]["fields"][field] = (
                        field if field != "Id" else "sf_id"
                    )
            if lookups:
                lookups.sort(key=field_sort)
                self.mapping[key]["lookups"] = {}
                for field in lookups:
                    # First, determine what manner of lookup we have here.
                    referenceTo = self.schema[obj][field]["referenceTo"]

                    if len(referenceTo) > 1:  # Polymorphic lookup
                        self.logger.warning(
                            f"Field {obj}.{field} is a polymorphic lookup, which is not supported"
                        )
                    elif referenceTo[0] == obj:  # Self-lookup
                        self.mapping[key]["lookups"][field] = {
                            "table": referenceTo[0],
                            "after": key,
                        }
                    elif stack.index(referenceTo[0]) > stack.index(
                        obj
                    ):  # Dependent lookup
                        self.mapping[key]["lookups"][field] = {
                            "table": referenceTo[0],
                            "after": f"Insert {referenceTo[0]}",
                        }
                    else:  # Regular lookup
                        self.mapping[key]["lookups"][field] = {"table": referenceTo[0]}

    def _split_dependencies(self, objs, dependencies):
        """Attempt to flatten the object network into a sequence of load operations."""
        stack = []
        objs_remaining = objs.copy()

        # The structure of `dependencies` is:
        # key = object, value = set of objects it references.

        # Iterate through our list of objects
        # For each object, if it is not dependent on any other objects, place it at the end of the stack.
        # Once an object is placed in the stack, remove dependencies to it (they're satisfied)
        while objs_remaining:
            objs_without_deps = [
                obj
                for obj in objs_remaining
                if obj not in dependencies or not dependencies[obj]
            ]

            if not objs_without_deps:
                self.logger.info(
                    "CumulusCI needs help to complete the mapping; the schema contains reference cycles and unresolved dependencies."
                )
                self.logger.info(f"Mapped objects: {', '.join(stack)}")
                self.logger.info("Remaining objects:")
                for obj in objs_remaining:
                    self.logger.info(obj)
                    for other_obj in dependencies[obj]:
                        self.logger.info(
                            f"   references {other_obj} via: {', '.join(dependencies[obj][other_obj])}"
                        )
                choice = click.prompt(
                    "Which object should we load first?",
                    type=click.Choice(list(objs_remaining)),
                    show_choices=True,
                )
                objs_without_deps = [choice]

            for obj in objs_without_deps:
                stack.append(obj)

                # Remove all dependencies on this object (they're satisfied)
                for other_obj in dependencies:
                    if obj in dependencies.get(other_obj):
                        del dependencies[other_obj][obj]

                # Remove this object from our remaining set.
                objs_remaining.remove(obj)

        return stack

    def _is_any_custom_api_name(self, api_name):
        """True if the entity name is custom (including any package)."""
        return api_name.endswith("__c")

    def _is_our_custom_api_name(self, api_name):
        """True if the entity name is custom and has our namespace prefix (if we have one)
        or if the entity does not have a namespace"""
        return self._is_any_custom_api_name(api_name) and (
            (
                self.options["namespace_prefix"]
                and api_name.startswith(self.options["namespace_prefix"])
            )
            or api_name.count("__") == 1
        )

    def _is_core_field(self, api_name):
        """True if this field is one that we should always include regardless
        of other settings or field configuration, such as Contact.FirstName.
        DB-level required fields don't need to be so handled."""

        return api_name in self.core_fields

    def _is_object_mappable(self, obj):
        """True if this object is one we can map, meaning it's an sObject and not
        some other kind of entity, it's not ignored, it's Bulk API compatible,
        and it's not in a hard-coded list of entities we can't currently handle."""

        return not any(
            [
                obj["name"] in self.options["ignore"],  # User-specified exclusions
                obj["name"].endswith(
                    "ChangeEvent"
                ),  # Change Data Capture entities (which get custom fields)
                obj["name"].endswith("__mdt"),  # Custom Metadata Types (MDAPI only)
                obj["name"].endswith("__e"),  # Platform Events
                obj["customSetting"],  # Not Bulk API compatible
                obj["name"]  # Objects we can't or shouldn't load/save
                in [
                    "User",
                    "Group",
                    "LookedUpFromActivity",
                    "OpenActivity",
                    "Task",
                    "Event",
                    "ActivityHistory",
                ],
            ]
        )

    def _is_field_mappable(self, obj, field):
        """True if this field is one we can map, meaning it's not ignored,
        it's createable by the Bulk API, it's not a deprecated field,
        and it's not a type of reference we can't handle without special
        configuration (self-lookup or reference to objects not included
        in this operation)."""
        return not any(
            [
                f"{obj}.{field['name']}" in self.options["ignore"],  # User-ignored list
                "(Deprecated)" in field["label"],  # Deprecated managed fields
                field["type"] == "base64",  # No Bulk API support for base64 blob fields
                not field["createable"],  # Non-writeable fields
                field["type"] == "reference"  # Outside lookups
                and not self._are_lookup_targets_in_operation(field),
            ]
        )

    def _is_required_field(self, field):
        """True if the field is either database-level required or a master-detail
        relationship field."""
        return (field["createable"] and not field["nillable"]) or (
            field["type"] == "reference" and field["relationshipOrder"] == 1
        )

    def _has_our_custom_fields(self, obj):
        """True if the object is owned by us or contains any field owned by us."""
        return any(
            [self._is_our_custom_api_name(field["name"]) for field in obj["fields"]]
        )

    def _are_lookup_targets_in_operation(self, field):
        """True if this lookup field aims at objects we are already including (all targets
        must match, although we don't provide actual support for polymorphism)."""
        return all([f in self.mapping_objects for f in field["referenceTo"]])

    def _is_lookup_to_included_object(self, field):
        """True if this field is a lookup and also references only objects we are
        already including."""
        return field["type"] == "reference" and self._are_lookup_targets_in_operation(
            field
        )



from robot.libraries.BuiltIn import BuiltIn
import functools
from cumulusci.core.utils import dictmerge

"""
This module supports managing multiple location strategies. It
works like this:

0. Locators are stored in a global LOCATORS dictionary. The keys
   are the locator prefixes, the values are dictionaries containing
   the locators

1. Libraries can register a dictionary of locators with a prefix
   (eg: register_locators("npsp", {...})). These get added to
   LOCATORS

2. Open Test Browser calls selenium's add_location_strategy for each
   registered set of locators. (Note: Location strategies cannot be added before a
   browser is open).

3. Keywords can use dot notation to refer to locators. A colon
   separates the prefix from the locator, and a locator from a
   comma-separated string of arguments

Example:

    from a keyword library:

    | from cumulusci.robotframework.locator_manager import register_locators
    |
    | register_locators("example", {"dialog": {"button": "//button[text()='{}']"}})

    in a test:

    | page should contain element  example:dialog.button:Save

    ... will result in "//button[text()='Save']" being passed
    to 'page should contain element'

"""

LOCATORS = {}


def register_locators(prefix, locators):
    """Register a strategy with a set of locators or a keyword

    If the prefix is already known, merge in the new locators.
    """
    if prefix in LOCATORS:
        BuiltIn().log(f"merging keywords for prefix {prefix}", "DEBUG")
        dictmerge(LOCATORS[prefix], locators)
    else:
        BuiltIn().log(f"registering keywords for prefix {prefix}", "DEBUG")
        LOCATORS[prefix] = locators


def add_location_strategies():
    """Call selenium's add_location_strategy keyword for all strategies"""

    # selenium throws an error if the location strategy already
    # exists, so we use a flag to make sure this code is called
    # only once.
    selenium = BuiltIn().get_library_instance("SeleniumLibrary")
    for (prefix, strategy) in LOCATORS.items():
        try:
            BuiltIn().log(f"adding location strategy for '{prefix}'", "DEBUG")
            if isinstance(strategy, dict):
                selenium.add_location_strategy(
                    prefix, functools.partial(locate_element, prefix)
                )
            else:
                # not a dict? Just pass it through to selenium as-is
                # so that this function can register normal keywords
                selenium.add_location_strategy(prefix, strategy)
        except Exception as e:
            BuiltIn().log(f"unable to register locators: {e}", "DEBUG")


def locate_element(prefix, parent, locator, tag, constraints):
    """This is the function called by SeleniumLibrary when a custom locator
    strategy is used (eg: cci:foo.bar). We pass an additional argument,
    prefix, so we know which set of locators to use.

    This tokenizes the locator and then does a lookup in the dictionary associated
    with the given prefix. If any arguments are present, they are applied with
    .format() before being used to find an element.
    """

    # Ideally we should call get_webelements (plural) and filter
    # the results based on the tag and constraints arguments, but
    # the documentation on those arguments is virtually nil and
    # SeleniumLibrary's filter mechanism is a private function. In
    # practice it probably won't matter <shrug>.
    selenium = BuiltIn().get_library_instance("SeleniumLibrary")
    loc = translate_locator(prefix, locator)
    BuiltIn().log(f"locate_element: '{prefix}:{locator}' => {loc}", "DEBUG")
    try:
        element = selenium.get_webelement(loc)
    except Exception:
        raise Exception(
            f"Element with locator '{prefix}:{locator}' not found\ntranslated: '{loc}'"
        )
    return element


def translate_locator(prefix, locator):
    """Return the translated locator

    This uses the passed-in prefix and locator to find the
    proper element in the LOCATORS dictionary, and then formats it
    with any arguments that were part of the locator.

    """

    if ":" in locator:
        (path, argstring) = locator.split(":", 1)
    else:
        path = locator
        argstring = ""

    loc = LOCATORS[prefix]
    breadcrumbs = []
    try:
        for key in path.split("."):
            breadcrumbs.append(key)
            loc = loc[key.strip()]

    except KeyError:
        breadcrumb_path = ".".join(breadcrumbs)
        raise KeyError(f"locator {prefix}:{breadcrumb_path} not found")

    if not isinstance(loc, str):
        raise TypeError(f"Expected locator to be of type string, but was {type(loc)}")

    try:
        # args is still a string, so split on ","
        # This means that arguments can't have commas in them, but I'm not sure
        # that will be a problem. If we find a case where it's a problem we can
        # do more sophisticated parsing.
        args = [arg.strip() for arg in argstring.split(",")] if argstring else []
        loc = loc.format(*args)
    except IndexError:
        raise Exception("Not enough arguments were supplied")
    return loc



from unittest import mock

from cumulusci.tasks.salesforce.GetInstalledPackages import GetInstalledPackages
from cumulusci.tasks.salesforce.tests.util import create_task


class TestGetInstalledPackages:
    @mock.patch(
        "cumulusci.salesforce_api.metadata.ApiRetrieveInstalledPackages.__call__"
    )
    def test_run_task(self, api):
        task = create_task(GetInstalledPackages)
        task()
        api.assert_called_once()
